"""django_mongo_coupons URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.8/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Add an import:  from blog import urls as blog_urls
    2. Add a URL to urlpatterns:  url(r'^blog/', include(blog_urls))
"""
from django.conf.urls import include, url

from django_mongo_coupons.views import CouponGenerationView, CampaignView

coupon_list = CouponGenerationView.as_view({
    'get': 'list',
    'post': 'create'
})
coupon_detail = CouponGenerationView.as_view({
    'get': 'retrieve',
    'put': 'update',
    'delete': 'destroy'
})

campaign_list = CampaignView.as_view({
    'get': 'list',
    'post': 'create'
})
campaign_detail = CampaignView.as_view({
    'get': 'retrieve',
    'put': 'update',
    'delete': 'destroy'
})


urlpatterns = [
    url(r'^coupons/$', coupon_list, name='coupon-list'),
    url(r'^coupons/(?P<code>\w+)/$', coupon_detail, name='coupon-detail'),
    url(r'^campaign/$', campaign_list, name='campaign-list'),
    url(r'^campaign/(?P<name>\w+)/$', campaign_detail, name='campaign-detail'),
    url(r'^check/(?P<name>\w+)/$', campaign_detail, name='campaign-detail'),
]


import random
from bson import ObjectId

from datetime import datetime

from django.conf import settings
from django.db import IntegrityError
from django.db import models
from django.dispatch import Signal
from django.utils.encoding import python_2_unicode_compatible
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
# from django_mongoengine import Document as Document
# from django_mongoengine import fields as dj
from django_mongoengine.queryset import QuerySetManager
from mongoengine import Document, fields
from django_mongo_coupons.coupon_settings import User

from django_mongo_coupons.exceptions import CouponAlreadyUsed


from .coupon_settings import (
    COUPON_TYPES,
    CODE_LENGTH,
    CODE_CHARS,
    SEGMENTED_CODES,
    SEGMENT_LENGTH,
    SEGMENT_SEPARATOR,
)


# try:
#     user_model = settings.AUTH_USER_MODEL
# except AttributeError:
#     from django.contrib.auth.models import User as user_model
redeem_done = Signal(providing_args=["coupon"])


class CouponManager(QuerySetManager):
    def create_coupon(self, coupon_type, value, users=[], valid_until=None, prefix="", campaign=None, user_limit=None,
                      usage_limit=None):
        coupon = self.create(
            value=value,
            code=Coupon.generate_code(prefix),
            type=coupon_type,
            valid_until=valid_until,
            campaign=campaign,
        )
        if user_limit is not None:  # otherwise use default value of model
            coupon.user_limit = user_limit
        if usage_limit is not None:  # otherwise use default value of model
            coupon.usage_limit = usage_limit

        try:
            coupon.save()
        except IntegrityError:
            # Try again with other code
            coupon = Coupon.objects.create_coupon(type, value, users, valid_until, prefix, campaign)
        if not isinstance(users, list):
            users = [users]
        for user in users:
            if user:
                CouponUser(user=user, coupon=coupon).save()
        return coupon

    def create_coupons(self, quantity, type, value, valid_until=None, prefix="", campaign=None):
        coupons = []
        for i in range(quantity):
            coupons.append(self.create_coupon(type, value, None, valid_until, prefix, campaign))
        return coupons

    def used(self):
        return self.exclude(users__redeemed_at__exists=True)

    def unused(self):
        return self.filter(users__redeemed_at__exists=False)

    def expired(self):
        return self.filter(valid_until__lt=timezone.now())

    def valid(self):
        return self.filer(Q(users__redeemed_at__exists=False) & Q(valid_until__gt=datetime.utcnow()))


# @python_2_unicode_compatible
class Coupon(Document):
    value = fields.IntField(verbose_name="Value", help_text=_("Arbitrary coupon value"))
    code = fields.StringField(required=False, verbose_name="Code", unique=True, max_length=30, null=True)
    max_discount = fields.IntField(required=False, verbose_name="Maximum discount", null=True)
    type = fields.StringField(verbose_name="Type", max_length=20, choices=COUPON_TYPES)
    user_limit = fields.IntField(verbose_name="User limit", default=1, min_value=0)
    usage_limit = fields.IntField(verbose_name="Usage limit per user", default=1, min_value=0)
    created_at = fields.DateTimeField(verbose_name="Created at", default=datetime.utcnow())
    valid_until = fields.DateTimeField(verbose_name="Valid until", blank=True, null=True,
                                       help_text="Leave empty for coupons that never expire")
    campaign = fields.ReferenceField('Campaign', verbose_name="Campaign", blank=True, null=True,
                                     related_name='coupons', dbref=True)
    kwargs = fields.DictField(required=False, null=True)

    objects = CouponManager()

    meta = {
        'collection': "coupons",
        'indexes': ['code', 'valid_until']
    }

    def __str__(self):
        return self.code

    def save(self, *args, **kwargs):
        if not self.code:
            self.code = Coupon.generate_code()
        super(Coupon, self).save(*args, **kwargs)

    def expired(self):
        return self.valid_until is not None and self.valid_until < timezone.now()

    @property
    def is_redeemed(self):
        """ Returns true is a coupon is redeemed (completely for all users) otherwise returns false. """
        return self.users.filter(
            redeemed_at__exists=True
        ).count() >= self.user_limit and self.user_limit is not 0

    @property
    def redeemed_at(self):
        # try:
        return CouponUser.objects.get(coupon=self, redeemed_at__ne=[]).sort('-redeemed_at').first().redeemed_at[-1]
        #     # return self.users.filter(redeemed_at__exists=True).order_by('redeemed_at').last().redeemed_at
        # except CouponUser.through.DoesNotExist:
        #     return None

    @classmethod
    def generate_code(cls, prefix="", segmented=SEGMENTED_CODES):
        code = "".join(random.choice(CODE_CHARS) for i in range(CODE_LENGTH))
        if segmented:
            code = SEGMENT_SEPARATOR.join([code[i:i + SEGMENT_LENGTH] for i in range(0, len(code), SEGMENT_LENGTH)])
            return prefix + code
        else:
            return prefix + code

    def redeem(self, user=None):
        if user:
            user = User.objects.get(id=user)
        try:
            coupon_user = CouponUser.objects.get(coupon=self,
                                                 user=user)
            print "coupon_user", coupon_user
        except CouponUser.DoesNotExist:
            try:  # silently fix unbouned or nulled coupon users
                coupon_user = CouponUser.objects.get(user__exists=False)
                coupon_user.user = user
                print "here1"
            except CouponUser.DoesNotExist:
                print "here2"
                coupon_user = CouponUser(coupon=self, user=user)
        if self.usage_limit and coupon_user.redeemed_at and len(coupon_user.redeemed_at) >= self.usage_limit:
            raise CouponAlreadyUsed
        coupon_user.redeemed_at.append(timezone.now())
        coupon_user.save()
        redeem_done.send(sender=self.__class__, coupon=self)
        
    def apply_coupon(self, amount):
        '''amount: amount to be paid'''
        if self.type == 'percentage':
            discount = amount * self.value / 100
            try:
                if self.max_discount and discount > self.max_discount:
                    discount = self.max_discount
            except AttributeError:
                pass
        else:
            discount = self.value
        amount = amount - discount
        return amount if amount > 0 else 0


@python_2_unicode_compatible
class Campaign(Document):
    name = fields.StringField(max_length=255, unique=True)
    description = fields.StringField(null=True)
    kwargs = fields.DictField(required=False, null=True)

    meta = {
        'collection': "campaign"
    }

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class CouponUser(Document):
    coupon = fields.ReferenceField(Coupon, dbref=True)
    user = fields.ReferenceField(User, dbref=True, null=True) #
    # , unique_with=coupon)
    redeemed_at = fields.ListField(fields.DateTimeField(verbose_name="Redeemed at", null=True))
    kwargs = fields.DictField(required=False, null=True)

    meta = {
        'collection': "coupon_user",
        'indexes': ['coupon', 'user', 'redeemed_at' ]
    }

    def __str__(self):
        return str(self.user)



from django.conf import settings
from django.http import Http404
from mongoengine import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_mongoengine.viewsets import ModelViewSet

from django_mongo_coupons.coupon_settings import User
from django_mongo_coupons.models import Coupon, Campaign, CouponUser
from django_mongo_coupons.serializer import CouponGenSerializer, CampaignSerializer


class CouponGenerationView(ModelViewSet):
    model = Coupon
    serializer_class = CouponGenSerializer
    lookup_field = 'code'
    queryset = Coupon.objects

    def get_queryset(self):
        return Coupon.objects.all()

    def retrieve(self, request, *args, **kwargs):
        amount = int(request.query_params.get('amount', None))
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        amount = instance.apply_coupon(amount)
        return_dict = serializer.data
        return_dict['amount'] = amount
        return Response(return_dict)


class CampaignView(ModelViewSet):
    model = Campaign
    serializer_class = CampaignSerializer
    queryset = Campaign.objects
    lookup_field = 'name'


    def get_queryset(self):
        return Campaign.objects.all()


class CouponView(APIView):

    def post(self, request):
        code = request.data['code']
        types = request.data['types']
        try:
            coupon = Coupon.objects.get(code=code)
        except Coupon.DoesNotExist:
            raise ValidationError(_("This code is not valid."))

        if request.user is None and coupon.user_limit is not 1:
            # coupons with can be used only once can be used without tracking the user, otherwise there is no chance
            # of excluding an unknown user from multiple usages.
            raise ValidationError(_(
                "The server must provide an user to this form to allow you to use this code. Maybe you need to sign in?"
            ))

        if coupon.is_redeemed:
            raise ValidationError(_("This code has already been used."))

        try:  # check if there is a user bound coupon existing
            user = User.objects.get(request.user)
            user_coupon = CouponUser.objects.get(user=user)
            if user_coupon.redeemed_at is not None:
                raise ValidationError(_("This code has already been used by your account."))
        except CouponUser.DoesNotExist:
            if coupon.user_limit is not 0:  # zero means no limit of user count
                # only user bound coupons left and you don't have one
                if coupon.user_limit is coupon.users.filter(user__isnull=False).count():
                    raise ValidationError(_("This code is not valid for your account."))
                if coupon.user_limit is coupon.users.filter(redeemed_at__isnull=False).count():  # all coupons redeemed
                    raise ValidationError(_("This code has already been used."))
        if types is not None and coupon.type not in types:
            raise ValidationError(_("This code is not meant to be used here."))
        if coupon.expired():
            raise ValidationError(_("This code is expired."))
        return code






import string

from django.conf import settings
from django_mongoengine.mongo_auth.models import MongoUser

COUPON_TYPES = getattr(settings, 'COUPONS_COUPON_TYPES', (
    ('monetary', 'Money based coupon'),
    ('percentage', 'Percentage discount'),
    ('virtual_currency', 'Virtual currency'),
))

CODE_LENGTH = getattr(settings, 'COUPONS_CODE_LENGTH', 15)

CODE_CHARS = getattr(settings, 'COUPONS_CODE_CHARS', string.ascii_letters+string.digits)

SEGMENTED_CODES = getattr(settings, 'COUPONS_SEGMENTED_CODES', False)
SEGMENT_LENGTH = getattr(settings, 'COUPONS_SEGMENT_LENGTH', 4)
SEGMENT_SEPARATOR = getattr(settings, 'COUPONS_SEGMENT_SEPARATOR', "-")

try:
    User = settings.COUPON_USER
except AttributeError:
    try:
        User = settings.AUTH_USER_MODEL
    except AttributeError:
        User = MongoUser


from rest_framework.status import HTTP_400_BAD_REQUEST


class BaseCustomException(Exception):
    """
    Used as the base class for all other exceptions
    Why? So that custom exceptions can be caught at the views
    """
    def _init__(self):
        super(BaseCustomException, self).__init__()
        self.code = 0000
        self.message = "An unknown BaseCustomException occurred"


class InvalidCouponCode(BaseCustomException):
    def __init__(self):
        super(InvalidCouponCode, self).__init__()
        self.code = HTTP_400_BAD_REQUEST
        self.message = "Invalid Coupon Code"


class CouponExpiredOrUsed(BaseCustomException):
    def __init__(self):
        super(CouponExpiredOrUsed, self).__init__()
        self.code = HTTP_400_BAD_REQUEST
        self.message = "The Coupon is either expired or used"


class CouponAlreadyUsed(BaseCustomException):
    def __init__(self):
        super(CouponAlreadyUsed, self).__init__()
        self.code = HTTP_400_BAD_REQUEST
        self.message = "The Coupon is already used"




from rest_framework_mongoengine.serializers import DocumentSerializer

from django_mongo_coupons.models import Coupon, Campaign, CouponUser


class CouponGenSerializer(DocumentSerializer):

    class Meta:
        model = Coupon
        fields = '__all__'


class CampaignSerializer(DocumentSerializer):
    class Meta:
        model = Campaign
        fields = '__all__'


class CouponUserSerializer(DocumentSerializer):
    class Meta:
        model = CouponUser
        fields = '__all__'




from mongoengine import ValidationError
from rest_framework_mongoengine.serializers import DocumentSerializer

from django_mongo_coupons.models import Coupon, CouponUser
from django.utils.translation import ugettext_lazy as _


class CouponSerializer(DocumentSerializer):

    class Meta:
        model = Coupon
        depth = 4
        fields = '__all__'

    def update(self, instance, validated_data):
        code = validated_data['code']
        try:
            coupon = Coupon.objects.get(code=code)
        except Coupon.DoesNotExist:
            raise ValidationError(_("This code is not valid."))
        self.coupon = coupon

        if self.user is None and coupon.user_limit is not 1:
            # coupons with can be used only once can be used without tracking the user, otherwise there is no chance
            # of excluding an unknown user from multiple usages.
            raise ValidationError(_(
                "The server must provide an user to this form to allow you to use this code. Maybe you need to sign in?"
            ))

        if coupon.is_redeemed:
            raise ValidationError(_("This code has already been used."))

        try:  # check if there is a user bound coupon existing
            user_coupon = coupon.users.get(user=self.user)
            if user_coupon.redeemed_at is not None:
                raise ValidationError(_("This code has already been used by your account."))
        except CouponUser.DoesNotExist:
            if coupon.user_limit is not 0:  # zero means no limit of user count
                # only user bound coupons left and you don't have one
                if coupon.user_limit is coupon.users.filter(user__isnull=False).count():
                    raise ValidationError(_("This code is not valid for your account."))
                if coupon.user_limit is coupon.users.filter(redeemed_at__isnull=False).count():  # all coupons redeemed
                    raise ValidationError(_("This code has already been used."))
        if self.types is not None and coupon.type not in self.types:
            raise ValidationError(_("This code is not meant to be used here."))
        if coupon.expired():
            raise ValidationError(_("This code is expired."))
        return code
